using LatticeGPU
using TOML
using TimerOutputs
using ArgParse
using CUDA


function fP_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] += (norm2(A11[b,r]) + norm2(A21[b,r]) + norm2(A31[b,r]) + norm2(A12[b,r]) + norm2(A22[b,r]) + norm2(A32[b,r]))/prod(lp.iL[1:3])
        end end end
    end
    return cf
end

function fA_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] -= (dot(A11[b,r],dmul(Gamma{4},A11[b,r])) + dot(A21[b,r],dmul(Gamma{4},A21[b,r])) + dot(A31[b,r],dmul(Gamma{4},A31[b,r]))
                      + dot(A12[b,r],dmul(Gamma{4},A12[b,r])) + dot(A22[b,r],dmul(Gamma{4},A22[b,r])) + dot(A32[b,r],dmul(Gamma{4},A32[b,r])))/prod(lp.iL[1:3])
        end end end
    end
    return cf
end

function gP_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] += (norm2(B11[b,r]) + norm2(B21[b,r]) + norm2(B31[b,r]) + norm2(B12[b,r]) + norm2(B22[b,r]) + norm2(B32[b,r]))/prod(lp.iL[1:3])
        end end end
    end
    return cf
end

function gA_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] += (dot(B11[b,r],dmul(Gamma{4},B11[b,r])) + dot(B21[b,r],dmul(Gamma{4},B21[b,r])) + dot(B31[b,r],dmul(Gamma{4},B31[b,r]))
                      + dot(B12[b,r],dmul(Gamma{4},B12[b,r])) + dot(B22[b,r],dmul(Gamma{4},B22[b,r])) + dot(B32[b,r],dmul(Gamma{4},B32[b,r])))/prod(lp.iL[1:3])
        end end end
    end
    return cf
end

function kV_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] += ( dot(A11[b,r],dmul(Gamma{6},imm(A12[b,r]))) + dot(A11[b,r],dmul(Gamma{7},-A12[b,r])) + dot(A11[b,r],dmul(Gamma{8}, imm(A11[b,r])))
                       +dot(A12[b,r],dmul(Gamma{6},imm(A11[b,r]))) + dot(A12[b,r],dmul(Gamma{7}, A11[b,r])) + dot(A12[b,r],dmul(Gamma{8},mimm(A12[b,r])))
                       +dot(A21[b,r],dmul(Gamma{6},imm(A22[b,r]))) + dot(A21[b,r],dmul(Gamma{7},-A22[b,r])) + dot(A21[b,r],dmul(Gamma{8}, imm(A21[b,r])))
                       +dot(A22[b,r],dmul(Gamma{6},imm(A21[b,r]))) + dot(A22[b,r],dmul(Gamma{7}, A21[b,r])) + dot(A22[b,r],dmul(Gamma{8},mimm(A22[b,r])))
                       +dot(A31[b,r],dmul(Gamma{6},imm(A32[b,r]))) + dot(A31[b,r],dmul(Gamma{7},-A32[b,r])) + dot(A31[b,r],dmul(Gamma{8}, imm(A31[b,r])))
                       +dot(A32[b,r],dmul(Gamma{6},imm(A31[b,r]))) + dot(A32[b,r],dmul(Gamma{7}, A31[b,r])) + dot(A32[b,r],dmul(Gamma{8},mimm(A32[b,r]))) )/(3*prod(lp.iL[1:3]))
        end end end
    end
    return cf
end

function lV_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] += ( dot(B11[b,r],dmul(Gamma{6},mimm(B12[b,r]))) + dot(B11[b,r],dmul(Gamma{7}, B12[b,r])) + dot(B11[b,r],dmul(Gamma{8},mimm(B11[b,r])))
                       +dot(B12[b,r],dmul(Gamma{6},mimm(B11[b,r]))) + dot(B12[b,r],dmul(Gamma{7},-B11[b,r])) + dot(B12[b,r],dmul(Gamma{8}, imm(B12[b,r])))
                       +dot(B21[b,r],dmul(Gamma{6},mimm(B22[b,r]))) + dot(B21[b,r],dmul(Gamma{7}, B22[b,r])) + dot(B21[b,r],dmul(Gamma{8},mimm(B21[b,r])))
                       +dot(B22[b,r],dmul(Gamma{6},mimm(B21[b,r]))) + dot(B22[b,r],dmul(Gamma{7},-B21[b,r])) + dot(B22[b,r],dmul(Gamma{8}, imm(B22[b,r])))
                       +dot(B31[b,r],dmul(Gamma{6},mimm(B32[b,r]))) + dot(B31[b,r],dmul(Gamma{7}, B32[b,r])) + dot(B31[b,r],dmul(Gamma{8},mimm(B31[b,r])))
                       +dot(B32[b,r],dmul(Gamma{6},mimm(B31[b,r]))) + dot(B32[b,r],dmul(Gamma{7},-B31[b,r])) + dot(B32[b,r],dmul(Gamma{8}, imm(B32[b,r]))) )/(3*prod(lp.iL[1:3]))
        end end end
    end
    return cf
end

function kT_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] +=im*(dot(A11[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A12[b,r])))) + dot(A11[b,r],dmul(Gamma{11},dmul(Gamma{5},-A12[b,r]))) + dot(A11[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A11[b,r]))))
                        +dot(A12[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A11[b,r])))) + dot(A12[b,r],dmul(Gamma{11},dmul(Gamma{5}, A11[b,r]))) + dot(A12[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A12[b,r]))))
                        +dot(A21[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A22[b,r])))) + dot(A21[b,r],dmul(Gamma{11},dmul(Gamma{5},-A22[b,r]))) + dot(A21[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A21[b,r]))))
                        +dot(A22[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A21[b,r])))) + dot(A22[b,r],dmul(Gamma{11},dmul(Gamma{5}, A21[b,r]))) + dot(A22[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A22[b,r]))))
                        +dot(A31[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A32[b,r])))) + dot(A31[b,r],dmul(Gamma{11},dmul(Gamma{5},-A32[b,r]))) + dot(A31[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(A31[b,r]))))
                        +dot(A32[b,r],dmul(Gamma{10},imm(dmul(Gamma{5},A31[b,r])))) + dot(A32[b,r],dmul(Gamma{11},dmul(Gamma{5}, A31[b,r]))) + dot(A32[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(A32[b,r])))) )/(3*prod(lp.iL[1:3]))
        end end end
    end
    return cf
end

function lT_fun()
    cf = Vector{Complex{Float64}}(undef,lp.iL[4])
    for t in 1:lp.iL[4]
        cf[t] = 0.0
        for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
            b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
            cf[t] +=-im*(dot(B11[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B12[b,r])))) + dot(B11[b,r],dmul(Gamma{11},dmul(Gamma{5}, B12[b,r]))) + dot(B11[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B11[b,r]))))
                         +dot(B12[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B11[b,r])))) + dot(B12[b,r],dmul(Gamma{11},dmul(Gamma{5},-B11[b,r]))) + dot(B12[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B12[b,r]))))
                         +dot(B21[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B22[b,r])))) + dot(B21[b,r],dmul(Gamma{11},dmul(Gamma{5}, B22[b,r]))) + dot(B21[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B21[b,r]))))
                         +dot(B22[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B21[b,r])))) + dot(B22[b,r],dmul(Gamma{11},dmul(Gamma{5},-B21[b,r]))) + dot(B22[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B22[b,r]))))
                         +dot(B31[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B32[b,r])))) + dot(B31[b,r],dmul(Gamma{11},dmul(Gamma{5}, B32[b,r]))) + dot(B31[b,r],dmul(Gamma{12},dmul(Gamma{5},mimm(B31[b,r]))))
                         +dot(B32[b,r],dmul(Gamma{10},mimm(dmul(Gamma{5},B31[b,r])))) + dot(B32[b,r],dmul(Gamma{11},dmul(Gamma{5},-B31[b,r]))) + dot(B32[b,r],dmul(Gamma{12},dmul(Gamma{5}, imm(B32[b,r])))) )/(3*prod(lp.iL[1:3]))
        end end end
    end
    return cf
end

f1_fun() = complex(norm2(C11) + norm2(C21) + norm2(C31) + norm2(C12) + norm2(C22) + norm2(C32))

k1_fun() = (dot(C11,dmul(Gamma{6},imm(C12))) + dot(C11,dmul(Gamma{7},-C12)) + dot(C11,dmul(Gamma{8}, imm(C11)))
            +dot(C12,dmul(Gamma{6},imm(C11))) + dot(C12,dmul(Gamma{7}, C11)) + dot(C12,dmul(Gamma{8},mimm(C12)))
            +dot(C21,dmul(Gamma{6},imm(C22))) + dot(C21,dmul(Gamma{7},-C22)) + dot(C21,dmul(Gamma{8}, imm(C21)))
            +dot(C22,dmul(Gamma{6},imm(C21))) + dot(C22,dmul(Gamma{7}, C21)) + dot(C22,dmul(Gamma{8},mimm(C22)))
            +dot(C31,dmul(Gamma{6},imm(C32))) + dot(C31,dmul(Gamma{7},-C32)) + dot(C31,dmul(Gamma{8}, imm(C31)))
            +dot(C32,dmul(Gamma{6},imm(C31))) + dot(C32,dmul(Gamma{7}, C31)) + dot(C32,dmul(Gamma{8},mimm(C32))) )/3



function compute_correlators()

    println(log_file,"\nComputing correlators")

    println(log_file,"\nComputing f1...\n")
    global f1 = f1_fun()
    println(log_file,"f1 =")
    println(log_file,f1)
    flush(log_file)

    println(log_file,"\nComputing k1...\n")
    global k1 = k1_fun()
    println(log_file,"k1 =")
    println(log_file,k1)
    flush(log_file)

    println(log_file,"\nComputing fP...\n")
    global fP = fP_fun()
    println(log_file,"fP =")
    for i in 1:lp.iL[4]
        println(log_file,fP[i])
    end
    flush(log_file)

    println(log_file,"\nComputing fA...\n")
    global fA = fA_fun()
    println(log_file,"fA =")
    for i in 1:lp.iL[4]
        println(log_file,fA[i])
    end
    flush(log_file)

    println(log_file,"\nComputing gP...\n")
    global gP = gP_fun()
    println(log_file,"gP =")
    for i in 1:lp.iL[4]
        println(log_file,gP[i])
    end
    flush(log_file)

    println(log_file,"\nComputing gA...\n")
    global gA = gA_fun()
    println(log_file,"gA =")
    for i in 1:lp.iL[4]
        println(log_file,gA[i])
    end
    flush(log_file)

    println(log_file,"\nComputing kV...\n")
    global kV = kV_fun()
    println(log_file,"kV =")
    for i in 1:lp.iL[4]
        println(log_file,kV[i])
    end
    flush(log_file)

    println(log_file,"\nComputing lV...\n")
    global lV = lV_fun()
    println(log_file,"lV =")
    for i in 1:lp.iL[4]
        println(log_file,lV[i])
    end
    flush(log_file)

    println(log_file,"\nComputing kT...\n")
    global kT = kT_fun()
    println(log_file,"kT =")
    for i in 1:lp.iL[4]
        println(log_file,kT[i])
    end
    flush(log_file)

    println(log_file,"\nComputing lT...\n")
    global lT = lT_fun()
    println(log_file,"lT =")
    for i in 1:lp.iL[4]
        println(log_file,lT[i])
    end
    flush(log_file)

end
