using LatticeGPU
using TOML
using TimerOutputs
using ArgParse
using CUDA

function load_fields()
    global U = vector_field(SU3{Float64}, lp)
    global U_CPU = Array(U)
    global psi = scalar_field(Spinor{4,SU3fund{Float64}},lp)
    global psi_CPU = Array(psi)

    global pp_density = Array{Float64}(undef,lp.bsz,lp.rsz));
    global ap_density = Array{Complex{Float64}}(undef,lp.bsz,lp.rsz));

    global pp_corr_t0 = fill(zero(Float64),(lp.iL[4],N_noise));
    global ap_corr_t0 = fill(zero(ComplexF64),(lp.iL[4],N_noise));

    global pp_corr_t = fill(zero(Float64),(lp.iL[4],N_noise,length(flow_times)));
    global ap_corr_t = fill(zero(ComplexF64),(lp.iL[4],N_noise,length(flow_times)));

    global Quark_cond = Array{Complex{Float64}}(undef,noise,length(flow_times));
end


function themalize()
    println(log_file,"Thermalizing")
    flush(log_file)

    for _ in 1:params["HMC"]["nth"]
        HMC!(U,intsch,lp,gp,ymws)
   end
end

function Gauge_update()

    println(log_file,"Performing "*string(params["HMC"]["dmeas"])*" HMC steps")

    flush(log_file)

    for _ in 1:params["HMC"]["dmeas"]
        _,acc = HMC!(U,intsch,lp,gp,ymws)
        acc ? nothing : println(log_file,"Rejected HCM step")
    end

    MCid += 1

    println(log_file,"Updated to Gauge configuration n"*string(MCid))

    Csw!(dws,U,gp,lp)
    U_CPU .= Array(U)

    flush(log_file)
    return nothing

end

"""
function two_pt()

Stores the two point function in the variables 'pp_corr_t0', 'ap_corr_t0', 'pp_corr_t' and 'ap_corr_t'
"""
function two_pt()

    for noi in 1:N_noise

        println(log_file,"Measuring 2pt noise source "*string(noi)*"...")
        flush(log_file)

        propagator!(psi,U, dpar, dws, lp,params["Measurements"]["maxiter"], params["Measurements"]["tolerance"],1)

        pp_density .= Array(norm2.(psi))
        ap_density .= Array(dot.(psi,dmul(Gamma{4},psi)))

        pp_corr_t0 .= zero(Float64)
        ap_corr_t0 .= zero(ComplexF64)
        for t in 1:lp.iL[4]
            for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
                b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
                pp_corr_t0[t,noi] += pp_density
                ap_corr_t0[t,noi] += ap_density
            end end end
        end



        delta_t = 0.0
        neps = int.eps_ini
        for fl in 1:length(flow_times)

            delta_t = flow_times[fl] - delta_t
            _,epslist = flw_adapt(U, psi, int, delta_t, neps, gp, dpar, lp, ymws, dws)
            neps = epslist[end]

            pp_density .= Array(norm2.(psi))
            ap_density .= Array(dot.(psi,dmul(Gamma{4},psi)))
            pp_corr_t .= zero(Float64)
            ap_corr_t .= zero(ComplexF64)
            for t in 1:lp.iL[4]
                for i in 1:lp.iL[1] for j in 1:lp.iL[2] for k in 1:lp.iL[3]
                    b,r = point_index(CartesianIndex{lp.ndim}((i,j,k,t)),lp)
                    pp_corr_t[t,noi,fl] += pp_density
                    ap_corr_t[t,noi,fl] += ap_density
                end end end
            end
        end

        copyto!(U,U_CPU)
    end

    println(log_file,"Finished measuring 2pt\n")
    flush(log_file)

    return nothing
end



function one_pt()

    copyto!(U,U_CPU)

    for noi in 1:N_noise
        for fl in 1:length(flow_times)

            println(log_file,"Measuring 1pt noise source "*string(noi)*"...")
            flush(log_file)

            pfrandomize!(psi,lp)

            backflow(psi, U, flow_times[fl], params["Measurements"]["Nsaves"], gp, dpar, lp, ymws, dws)

            psi_CPU .= Array(psi)

            dws.sp .= dmul.(Gamma{5},psi)
            g5Dw!(psi,U,dws.sp,dpar,dws,lp)

            CG!(psi,U,DwdagDw!,dpar,lp,dws,params["Measurements"]["maxiter"], params["Measurements"]["tolerance"])

            copyto!(dws.sp,psi_CPU)
            Quark_cond[noi,fl] = sum(dot.(dws.sp,psi) - params["Fermion"]["Cfl"].*norm2.(dws.sp))

            copyto!(U,U_CPU)

        end
    end
    println(log_file,"Finished measuring 1pt\n")

    return nothing
end

function gauge_meas()

    println(log_file,"\nGauge meassurements:\n")
    println(log_file,"Plaquette: ", plaquette(U, lp, gp, ymws))
    println(log_file,"Topological charge: ", Qtop(U, gp, lp, ymws))
    sfc = sfcoupling(U, lp, gp, ymws)
    flush(log_file)

    return nothing
end
